#
# Copyright (c) 2008 Tilman Sauerbeck (tilman at code-monkey de)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

class IOFile < File
end

class Integer
	def to_hex
		"0x%x" % self
	end
end

module Mvi
	class File
		attr_reader :header, :items

		# for most minimovies
		MAGIC = 0x800

		# for some minimovies
		#MAGIC = 0x1000

		# for films
		#MAGIC = 0xf000 # d02.mvi
		#MAGIC = 0x3800 # d03.mvi
		#MAGIC = 0x5800

		def initialize(filename)
			IOFile.open(filename, "rb") do |f|
				# read the 48 byte header
				buf = f.read(0x30)

				@header = Header.new(buf)
				@items = []

				# after the header there's 16 bytes that have information
				# about the chunks.
				@header.num_entries.times do
					buf = f.read(0x10)

					@items << Item.new(buf)
				end

				# seek to the beginning of the chunk payload
				f.seek(MAGIC)

				# and read the payload
				@items.each_with_index do |item, i|
					item.read_body(f)
				end
			end
		end

		def dump(filename, frame = :all)
			f = IOFile.open(filename, "wb")

			@header.dump(f)

			if frame == :all
				@items.each do |item|
					item.dump_header(f)
				end

				f.seek(MAGIC)

				@items.each do |item|
					item.dump_body(f)
				end
			else
				# the first item will always have to be stored
				@items.first.dump_header(f)

				(@items.length - 1).times do
					@items[frame + 1].dump_header(f)
				end

				f.seek(MAGIC)

				@items.first.dump_body(f)

				(@items.length - 1).times do
					@items[frame + 1].dump_body(f)
				end
			end
		ensure
			f.close
		end
	end

	class Header
		attr_reader :video_width, :video_height, :num_entries, :fps,
		            :audio_num_channels

		def initialize(buf)
			items = buf.unpack("VVVVVVVVVVVV")

			@major = items.shift
			@minor = items.shift

			@video_width = items.shift
			@video_height = items.shift

			@num_entries = items.shift
			@num_audio = items.shift
			@num_video = items.shift

			@fps = items.shift

			@has_audio = items.shift

			@audio_num_channels = items.shift
			@audio_samplerate = items.shift
			@audio_bits = items.shift
		end

		def has_audio?
			!@has_audio.zero?
		end

		def dump(f)
			f.write([
				@major,
				@minor,

				@video_width,
				@video_height,

				@num_entries,
				@num_audio,
				@num_video,

				@fps,

				@has_audio,

				@audio_num_channels,
				@audio_samplerate,
				@audio_bits,
			].pack("VVVVVVVVVVVV"))
		end
	end

	class Item
		attr_reader :channel, :length, :offset, :body

		def initialize(buf)
			items = buf.unpack("VVVV")

			@channel = items.shift
			@type = items.shift
			@length = items.shift
			@offset = items.shift

			@body = nil
		end

		# FIXME: needs a rename
		def type
			case @type
			when 0x00
				:colormap
			when 0x01
				:video
			when 0x02
				:audio
			end
		end

		def read_body(file)
			@body = file.read(@length)

			if @body.length != @length
				raise IOError, "cannot read full length. did you set up MAGIC correctly?"
			end
		end

		def dump_header(f)
			f.write([
				@channel,
				@type,
				@length,
				@offset
			].pack("VVVV"))
		end

		def dump_body(f)
			f.write(@body)
		end

		def decode
			FrameDecoder.new.decode(@body)
		end
	end

	class FrameDecoder
		INVALID_ENTRY = 0x02

		class HistoryEntry
			attr_accessor :parent, :color, :count

			def initialize(parent = INVALID_ENTRY,
			               color = 0x00, count = 0x00)
				@parent = parent
				@color = color
				@count = count
			end
		end

		def initialize
			@decoded = []
			@decoded_index = 0

			@history = []

			0.upto(255) do |i|
				@history[0x03 + i] =
					HistoryEntry.new(INVALID_ENTRY, i, 1)
			end

			@val = nil
			@last_read = nil
			@last_stored = nil
			@mask = nil

			@input_offset = 0x08 # pixel data starts here
		end

		def decode(buf)
			@buf = buf

			masks = [
				0x1ff0, 0x3ff0, 0x7ff0, 0xfff0, 0x1fff0, 0x3fff0, 0x7fff0
			]

			current_mask_index = 0
			current_loop_no = 1

			@mask = masks[current_mask_index]

			decode_start

			loop do
				case decode_loop(current_loop_no)
				when :done
					puts "we're done. thank god. #{current_loop_no}"
					return @decoded
				when :restart_with_different_mask
					current_mask_index += 1
					@mask = masks[current_mask_index]
					current_loop_no += 1
					puts "switching to loop #{current_loop_no}"
					puts "switching to mask #{@mask.to_hex}"
				when :decode_start
					# not handled yet. maybe not even necessary?
					raise "decode_start"
					return nil
				else
					# go on looping
				end
			end

			@decoded
		end

		def decode_loop(n)
			case n
			when 1
				decode_loop1
			when 2
				decode_loop2
			when 3
				decode_loop3
			when 4
				decode_loop4
			when 5
				decode_loop5
			when 6
				decode_loop6
			when 7
				decode_loop7
			end
		end

		def decode_start
			# first thingie
			@val = read_uint16_be(@input_offset)
			@val >>= 3
			@val &= @mask
			@val /= 0x10

			# get the color from the lookup table
			color = @history[@val].color

			@decoded[@decoded_index] = color
			@decoded_index += 1

			@last_color = color
			@last_stored = @val

			@is_first = true
		end

		def dump_reg(sym)
			puts "#{sym} #{instance_variable_get(sym).to_hex}"
		end

		def decode_loop1
			unless @is_first
				# block 0
				@val = read_uint16_be(@input_offset)
				@val >>= 3

				pre_decode
				return recover(0x02) if @val <= INVALID_ENTRY
				decode_macro
			end

			# block 1
			@val = read_uint16_be(@input_offset + 1) >> 2

			pre_decode
			return recover(0x03) if @val <= INVALID_ENTRY
			decode_macro

			# block 2
			@val = read_uint16_be(@input_offset + 2) >> 1

			pre_decode
			return recover(0x04) if @val <= INVALID_ENTRY
			decode_macro

			# block 3
			@val = read_uint16_be(@input_offset + 3)

			pre_decode
			return recover(0x05) if @val <= INVALID_ENTRY
			decode_macro

			# block 4
			@val = read_uint16_be(@input_offset + 4) << 1

			pre_decode
			return recover(0x06) if @val <= INVALID_ENTRY
			decode_macro

			# block 5
			@val = read_uint16_be(@input_offset + 5) << 2

			pre_decode
			return recover(0x07) if @val <= INVALID_ENTRY
			decode_macro

			# block 6

			@val = read_uint16_be(@input_offset + 6) << 3

			pre_decode
			return recover(0x08) if @val <= INVALID_ENTRY
			decode_macro

			# block 7
			@val = read_uint16_be(@input_offset + 7) << 4

			pre_decode
			return recover(0x09) if @val <= INVALID_ENTRY
			decode_macro

			@input_offset += 9
			@is_first = false
		end

		def decode_loop2
			unless @is_first
				# block 0
				@val = read_uint16_be(@input_offset) >> 2

				pre_decode
				return recover(0x02) if @val <= INVALID_ENTRY
				decode_macro
			end

			# block 1
			@val = read_uint16_be(@input_offset + 1)

			pre_decode
			return recover(0x03) if @val <= INVALID_ENTRY
			decode_macro

			# block 2
			@val = read_uint16_be(@input_offset + 2) << 2

			pre_decode
			return recover(0x04) if @val <= INVALID_ENTRY
			decode_macro

			# block 3
			@val = read_uint16_be(@input_offset + 3) << 4

			pre_decode
			return recover(0x05) if @val <= INVALID_ENTRY
			decode_macro

			@input_offset += 5
			@is_first = false
		end

		def decode_loop3
			unless @is_first
				# block 0
				@val = read_uint16_be(@input_offset) >> 1

				pre_decode
				return recover(0x02) if @val <= INVALID_ENTRY
				decode_macro
			end

			# block 1
			@val = read_uint16_be(@input_offset + 1) << 2

			pre_decode
			return recover(0x03) if @val <= INVALID_ENTRY
			decode_macro

			# block 2
			@val = read_uint16_be(@input_offset + 2)
			tmp = @buf[@input_offset + 1, 4].unpack("V").first

			@val <<= 5
			@val |= tmp >> (32 - 5)

			pre_decode
			return recover(0x05) if @val <= INVALID_ENTRY
			decode_macro

			# block 3
			@val = read_uint16_be(@input_offset + 4)

			pre_decode
			return recover(0x06) if @val <= INVALID_ENTRY
			decode_macro

			# block 4
			@val = read_uint16_be(@input_offset + 5) << 3

			pre_decode
			return recover(0x07) if @val <= INVALID_ENTRY
			decode_macro

			# block 5
			@val = read_uint16_be(@input_offset + 6)
			tmp = @buf[@input_offset + 5, 4].unpack("V").first

			@val <<= 6
			@val |= tmp >> (32 - 6)

			pre_decode
			return recover(0x09) if @val <= INVALID_ENTRY
			decode_macro

			# block 6
			@val = read_uint16_be(@input_offset + 8) << 1

			pre_decode
			return recover(0x0a) if @val <= INVALID_ENTRY
			decode_macro

			# block 7
			@val = read_uint16_be(@input_offset + 9) << 4

			pre_decode
			return recover(0x0b) if @val <= INVALID_ENTRY
			decode_macro

			@input_offset += 0x0b
			@is_first = false
		end

		def decode_loop4
			unless @is_first
				# block 0
				@val = read_uint16_be(@input_offset)

				pre_decode
				return recover(0x02) if @val <= INVALID_ENTRY
				decode_macro
			end

			# block one
			@val = read_uint16_be(@input_offset + 1) << 4

			pre_decode
			return recover(0x03) if @val <= INVALID_ENTRY
			decode_macro

			@input_offset += 0x03
			@is_first = false
		end

		def decode_loop5
			unless @is_first
				# block 0
				@val = read_uint16_be(@input_offset);
				@val <<= 1

				pre_decode
				return recover(0x02) if @val <= INVALID_ENTRY
				decode_macro
			end

			# block 1
			@val = read_uint16_be(@input_offset + 1)
			tmp = @buf[@input_offset, 4].unpack("V").first

			@val <<= 6
			@val |= tmp >> (32 - 6)

			pre_decode
			return recover(0x04) if @val <= INVALID_ENTRY
			decode_macro

			# block 2
			@val = read_uint16_be(@input_offset + 3) << 3

			pre_decode
			return recover(0x05) if @val <= INVALID_ENTRY
			decode_macro

			# block 3
			@val = read_uint16_be(@input_offset + 4)
			tmp = @buf[@input_offset + 3, 4].unpack("V").first

			@val <<= 8
			@val |= tmp >> (32 - 8)

			pre_decode
			return recover(0x07) if @val <= INVALID_ENTRY
			decode_macro

			# block 4
			@val = read_uint16_be(@input_offset + 6)
			tmp = @buf[@input_offset + 5, 4].unpack("V").first

			@val <<= 5
			@val |= tmp >> (32 - 5)

			pre_decode
			return recover(0x09) if @val <= INVALID_ENTRY
			decode_macro

			# block 5
			@val = read_uint16_be(@input_offset + 8) << 2

			pre_decode
			return recover(0x0a) if @val <= INVALID_ENTRY
			decode_macro

			# block 6
			@val = read_uint16_be(@input_offset + 9)
			tmp = @buf[@input_offset + 8, 4].unpack("V").first

			@val <<= 7
			@val |= tmp >> (32 - 7)

			pre_decode
			return recover(0x0c) if @val <= INVALID_ENTRY
			decode_macro

			# block 7
			@val = read_uint16_be(@input_offset + 0x0b) << 4

			pre_decode
			return recover(0x0d) if @val <= INVALID_ENTRY
			decode_macro

			@input_offset += 0x0d
			@is_first = false
		end

		def decode_loop6
			unless @is_first
				# block 0
				@val = read_uint16_be(@input_offset) << 2

				pre_decode
				return recover(0x02) if @val <= INVALID_ENTRY
				decode_macro
			end

			# block one
			@val = read_uint16_be(@input_offset + 1)
			tmp = @buf[@input_offset, 4].unpack("V").first

			@val <<= 8
			@val |= tmp >> (32 - 8)

			pre_decode
			return recover(0x04) if @val <= INVALID_ENTRY
			decode_macro

			# block 2
			@val = read_uint16_be(@input_offset + 3)

			tmp = @buf[@input_offset + 2, 4].unpack("V").first

			@val <<= 6
			@val |= tmp >> (32 - 6)

			pre_decode
			return recover(0x06) if @val <= INVALID_ENTRY
			decode_macro

			# block 3
			@val = read_uint16_be(@input_offset + 5) << 4

			pre_decode
			return recover(0x07) if @val <= INVALID_ENTRY
			decode_macro

			@input_offset += 0x07
			@is_first = false
		end

		def decode_loop7
			unless @is_first
				# block 0
				@val = read_uint16_be(@input_offset) << 3

				pre_decode
				return recover(0x02) if @val <= INVALID_ENTRY
				decode_macro
			end

			# block one
			@val = read_uint16_be(@input_offset + 1)
			tmp = @buf[@input_offset, 4].unpack("V").first

			@val <<= 10
			@val |= tmp >> (32 - 10)

			pre_decode
			return recover(0x04) if @val <= INVALID_ENTRY
			decode_macro

			# block 2
			@val = read_uint16_be(@input_offset + 3)
			tmp = @buf[@input_offset + 2, 4].unpack("V").first

			@val <<= 9
			@val |= tmp >> (32 - 9)

			pre_decode
			return recover(0x06) if @val <= INVALID_ENTRY
			decode_macro

			# block 3
			@val = read_uint16_be(@input_offset + 5)
			tmp = @buf[@input_offset + 4, 4].unpack("V").first

			@val <<= 8
			@val |= tmp >> (32 - 8)

			pre_decode
			return recover(0x08) if @val <= INVALID_ENTRY
			decode_macro

			# block 4
			@val = read_uint16_be(@input_offset + 7)
			tmp = @buf[@input_offset + 6, 4].unpack("V").first

			@val <<= 7
			@val |= tmp >> (32 - 7)

			pre_decode
			return recover(0x0a) if @val <= INVALID_ENTRY
			decode_macro

			# block 5
			@val = read_uint16_be(@input_offset + 0x09)
			tmp = @buf[@input_offset + 0x08, 4].unpack("V").first

			@val <<= 0x06
			@val |= tmp >> (32 - 0x06)

			pre_decode
			return recover(0x0c) if @val <= INVALID_ENTRY
			decode_macro

			# block 6
			@val = read_uint16_be(@input_offset + 0x0b)
			tmp = @buf[@input_offset + 0x0a, 4].unpack("V").first

			@val <<= 0x05
			@val |= tmp >> (32 - 0x05)

			pre_decode
			return recover(0x0e) if @val <= INVALID_ENTRY
			decode_macro

			# block 7
			@val = read_uint16_be(@input_offset + 0x0d) << 4

			pre_decode
			return recover(0x0f) if @val <= INVALID_ENTRY
			decode_macro

			@input_offset += 0x0f
			@is_first = false
		end

		# read a 16 bit unsigned integer in big endian format from offset
		def read_uint16_be(offset)
			v = @buf[offset + 0]
			v <<= 8
			v |= @buf[offset + 1]

			v
		end

		def pre_decode
			@val &= @mask
			@val /= 0x10

			@last_read = @val
		end

		def decode_macro
			if @val >= @history.length
				@val = @last_stored

				count = @history[@val].count
				@decoded[@decoded_index + count] = @last_color
			end

			@decoded_index += @history[@val].count
			color = nil

			loop do
				color = @history[@val].color

				@decoded_index -= 1
				@decoded[@decoded_index] = color

				@val = @history[@val].parent

				break unless @val != INVALID_ENTRY
			end

			entry = HistoryEntry.new
			@history << entry

			entry.color = color
			entry.parent = @last_stored
			entry.count = @history[entry.parent].count + 1

			@last_color = entry.color

			@last_stored = @last_read

			@decoded_index += @history[@last_stored].count
		end

		def recover(n)
			# skip the bytes that we've read so far
			@input_offset += n

			if @mask != 0x7fff0 && @val.zero?
				:restart_with_different_mask
			elsif @val == 0x2
				:done
			else
				:decode_start
			end
		end
	end
end
